"""
Created on 29 Jun 2015

@author: @willu47

This module provides the basic infrastructure for plotting charts for the
Method of Morris results

The procedures should build upon and return an axes instance::

    import matplotlib.pyplot as plt
    Si = morris.analyze(problem, param_values, Y, conf_level=0.95,
                        print_to_console=False, num_levels=10)

    # set plot style etc.
    fig, ax = plt.subplots(1)
    p = SALib.plotting.morris.horizontal_bar_plot(ax, Si, {'marker':'x'})
    p.show()
"""
import numpy as np


def _sort_Si(Si, key, sortby="mu_star"):
    return np.array([Si[key][x] for x in np.argsort(Si[sortby])])


def _sort_Si_by_index(Si, key, index):
    return np.array([Si[key][x] for x in index])


def horizontal_bar_plot(ax, Si, opts=None, sortby="mu_star", unit=""):
    """Updates a matplotlib axes instance with a horizontal bar plot
    of mu_star, with error bars representing mu_star_conf.
    """
    assert sortby in ["mu_star", "mu_star_conf", "sigma", "mu"]

    if opts is None:
        opts = {}

    # Sort all the plotted elements by mu_star (or optionally another
    # metric)
    names_sorted = _sort_Si(Si, "names", sortby)
    mu_star_sorted = _sort_Si(Si, "mu_star", sortby)
    mu_star_conf_sorted = _sort_Si(Si, "mu_star_conf", sortby)

    # Plot horizontal barchart
    y_pos = np.arange(len(mu_star_sorted))
    plot_names = names_sorted

    out = ax.barh(
        y_pos,
        mu_star_sorted,
        xerr=mu_star_conf_sorted,
        align="center",
        ecolor="black",
        **opts
    )

    ax.set_yticks(y_pos)
    ax.set_yticklabels(plot_names)
    ax.set_xlabel(r"$\mu^\star$" + unit)

    ax.set_ylim(min(y_pos) - 1, max(y_pos) + 1)

    return out


def covariance_plot(ax, Si, opts=None, unit=""):
    """Plots mu* against sigma or the 95% confidence interval"""
    if opts is None:
        opts = {}

    if Si["sigma"] is not None:
        # sigma is not present if using morris groups
        y = Si["sigma"]
        out = ax.scatter(Si["mu_star"], y, c="k", marker="o", **opts)
        ax.set_ylabel(r"$\sigma$")

        ax.set_xlim(
            0,
        )
        ax.set_ylim(
            0,
        )

        x_axis_bounds = np.array(ax.get_xlim())

        (line1,) = ax.plot(x_axis_bounds, x_axis_bounds, "k-")
        (line2,) = ax.plot(x_axis_bounds, 0.5 * x_axis_bounds, "k--")
        (line3,) = ax.plot(x_axis_bounds, 0.1 * x_axis_bounds, "k-.")

        ax.legend(
            (line1, line2, line3),
            (
                r"$\sigma / \mu^{\star} = 1.0$",
                r"$\sigma / \mu^{\star} = 0.5$",
                r"$\sigma / \mu^{\star} = 0.1$",
            ),
            loc="best",
        )

    else:
        y = Si["mu_star_conf"]
        out = ax.scatter(Si["mu_star"], y, c="k", marker="o", **opts)
        ax.set_ylabel(r"$95\% CI$")

    ax.set_xlabel(r"$\mu^\star$ " + unit)
    ax.set_ylim(
        0 - (0.01 * np.array(ax.get_ylim()[1])),
    )

    return out


def sample_histograms(fig, input_sample, problem, opts=None):
    """Plots a set of subplots of histograms of the input sample"""
    if opts is None:
        opts = {}

    num_vars = problem["num_vars"]
    names = problem["names"]

    framing = 101 + (num_vars * 10)

    # Find number of levels
    num_levels = len(set(input_sample[:, 1]))

    out = []

    for variable in range(num_vars):
        ax = fig.add_subplot(framing + variable)
        out.append(
            ax.hist(
                input_sample[:, variable],
                bins=num_levels,
                density=False,
                label=None,
                **opts
            )
        )

        ax.set_title("%s" % (names[variable]))
        ax.tick_params(
            axis="x",  # changes apply to the x-axis
            which="both",  # both major and minor ticks are affected
            bottom="off",  # ticks along the bottom edge are off
            top="off",  # ticks along the top edge are off
            labelbottom="off",
        )  # labels along the bottom edge off)
        if variable > 0:
            ax.tick_params(
                axis="y",  # changes apply to the y-axis
                which="both",  # both major and minor ticks affected
                labelleft="off",
            )  # labels along the left edge off)

    return out


if __name__ == "__main__":
    pass



import math
import numpy as np
from scipy.stats import norm

from . import common_args
from ..util import read_param_file, ResultDict


def analyze(
    problem,
    Y,
    M=4,
    num_resamples=100,
    conf_level=0.95,
    print_to_console=False,
    seed=None,
):
    """Perform extended Fourier Amplitude Sensitivity Test on model outputs.

    Returns a dictionary with keys 'S1' and 'ST', where each entry is a list of
    size D (the number of parameters) containing the indices in the same order
    as the parameter file.

    Notes
    -----
    Compatible with:
        `fast_sampler` : :func:`SALib.sample.fast_sampler.sample`

    Examples
    --------
        >>> X = fast_sampler.sample(problem, 1000)
        >>> Y = Ishigami.evaluate(X)
        >>> Si = fast.analyze(problem, Y, print_to_console=False)

    Parameters
    ----------
    problem : dict
        The problem definition
    Y : numpy.array
        A NumPy array containing the model outputs
    M : int
        The interference parameter, i.e., the number of harmonics to sum in
        the Fourier series decomposition (default 4)
    print_to_console : bool
        Print results directly to console (default False)
    seed : int
        Seed to generate a random number

    References
    ----------
    1. Cukier, R. I., C. M. Fortuin, K. E. Shuler, A. G. Petschek, and
       J. H. Schaibly (1973).
       Study of the sensitivity of coupled reaction systems to
       uncertainties in rate coefficients.
       J. Chem. Phys., 59(8):3873-3878
       doi:10.1063/1.1680571

    2. Saltelli, A., S. Tarantola, and K. P.-S. Chan (1999).  
       A Quantitative Model-Independent Method for Global Sensitivity Analysis
       of Model Output.
       Technometrics, 41(1):39-56,
       doi:10.1080/00401706.1999.10485594.

    3. Pujol, G. (2006)
       fast99 - R `sensitivity` package
       https://github.com/cran/sensitivity/blob/master/R/fast99.R
    """
    if seed:
        np.random.seed(seed)

    D = problem["num_vars"]

    if Y.size % (D) == 0:
        N = int(Y.size / D)
    else:
        msg = """
        Error: Number of samples in model output file must be a multiple of D,
        where D is the number of parameters.
        """
        raise ValueError(msg)

    # Recreate the vector omega used in the sampling
    omega_0 = math.floor((N - 1) / (2 * M))

    # Calculate and Output the First and Total Order Values
    Si = ResultDict((k, [None] * D) for k in ["S1", "ST", "S1_conf", "ST_conf"])
    Si["names"] = problem["names"]
    for i in range(D):
        z = np.arange(i * N, (i + 1) * N)

        Y_l = Y[z]

        S1, ST = compute_orders(Y_l, N, M, omega_0)
        Si["S1"][i] = S1
        Si["ST"][i] = ST

        S1_d_conf, ST_d_conf = bootstrap(Y_l, M, num_resamples, conf_level)
        Si["S1_conf"][i] = S1_d_conf
        Si["ST_conf"][i] = ST_d_conf

    if print_to_console:
        print(Si.to_df())

    return Si


def compute_orders(outputs: np.ndarray, N: int, M: int, omega: int):
    f = np.fft.fft(outputs)
    Sp = np.power(np.absolute(f[np.arange(1, math.ceil(N / 2))]) / N, 2)

    V = 2.0 * np.sum(Sp)

    # Calculate first and total order
    D1 = 2.0 * np.sum(Sp[np.arange(1, M + 1) * omega - 1])
    Dt = 2.0 * np.sum(Sp[np.arange(math.floor(omega / 2.0))])

    return (D1 / V), (1.0 - Dt / V)


def bootstrap(Y: np.ndarray, M: int, resamples: int, conf_level: float):
    """Compute CIs.

    Infers ``N`` from results of sub-sample ``Y`` and re-estimates omega (ω)
    for the above ``N``.
    """
    # Use half of available data each time
    T_data = Y.shape[0]
    n_size = math.ceil(T_data * 0.5)

    res_S1 = np.zeros(resamples)
    res_ST = np.zeros(resamples)
    for i in range(resamples):
        sample_idx = np.random.choice(T_data, replace=True, size=n_size)
        Y_rs = Y[sample_idx]

        N = len(Y_rs)
        omega = math.floor((N - 1) / (2 * M))

        S1, ST = compute_orders(Y_rs, N, M, omega)
        res_S1[i] = S1
        res_ST[i] = ST

    bnd = norm.ppf(0.5 + conf_level / 2.0)
    S1_conf = bnd * res_S1.std(ddof=1)
    ST_conf = bnd * res_ST.std(ddof=1)
    return S1_conf, ST_conf


# No additional arguments required for FAST
def cli_parse(parser):
    """Add method specific options to CLI parser.

    Parameters
    ----------
    parser : argparse object

    Returns
    -------
    Updated argparse object
    """
    parser.add_argument(
        "-M", "--M", type=int, required=False, default=4, help="Inference parameter"
    )
    parser.add_argument(
        "-r",
        "--resamples",
        type=int,
        required=False,
        default=100,
        help="Number of bootstrap resamples for Sobol " "confidence intervals",
    )

    return parser


def cli_action(args):
    problem = read_param_file(args.paramfile)
    Y = np.loadtxt(
        args.model_output_file, delimiter=args.delimiter, usecols=(args.column,)
    )

    analyze(
        problem,
        Y,
        M=args.M,
        num_resamples=args.resamples,
        print_to_console=True,
        seed=args.seed,
    )


if __name__ == "__main__":
    common_args.run_cli(cli_parse, cli_action)



import math

import numpy as np

from . import common_args
from ..util import scale_samples, read_param_file


def sample(problem, N, M=4, seed=None):
    """Generate model inputs for extended Fourier Amplitude Sensitivity Test.

    Returns a NumPy matrix containing the model inputs required by the extended
    Fourier Amplitude sensitivity test.  The resulting matrix contains N * D
    rows and D columns, where D is the number of parameters.

    The samples generated are intended to be used by
    :func:`SALib.analyze.fast.analyze`.

    Parameters
    ----------
    problem : dict
        The problem definition
    N : int
        The number of samples to generate
    M : int
        The interference parameter, i.e., the number of harmonics to sum in the
        Fourier series decomposition (default 4)
    seed : int
        Seed to generate a random number

    References
    ----------
    1. Cukier, R.I., Fortuin, C.M., Shuler, K.E., Petschek, A.G.,
       Schaibly, J.H., 1973.
       Study of the sensitivity of coupled reaction systems to
       uncertainties in rate coefficients. I theory.
       Journal of Chemical Physics 59, 3873-3878.
       https://doi.org/10.1063/1.1680571

    2. Saltelli, A., S. Tarantola, and K. P.-S. Chan (1999).
       A Quantitative Model-Independent Method for Global Sensitivity Analysis
       of Model Output.
       Technometrics, 41(1):39-56,
       doi:10.1080/00401706.1999.10485594.
    """
    if seed:
        np.random.seed(seed)

    if N <= 4 * M**2:
        raise ValueError(
            """
        Sample size N > 4M^2 is required. M=4 by default."""
        )

    D = problem["num_vars"]

    omega = np.zeros([D])
    omega[0] = math.floor((N - 1) / (2 * M))
    m = math.floor(omega[0] / (2 * M))

    if m >= (D - 1):
        omega[1:] = np.floor(np.linspace(1, m, D - 1))
    else:
        omega[1:] = np.arange(D - 1) % m + 1

    # Discretization of the frequency space, s
    s = (2 * math.pi / N) * np.arange(N)

    # Transformation to get points in the X space
    X = np.zeros([N * D, D])
    omega2 = np.zeros([D])

    for i in range(D):
        omega2[i] = omega[0]
        idx = list(range(i)) + list(range(i + 1, D))
        omega2[idx] = omega[1:]
        z = range(i * N, (i + 1) * N)

        # random phase shift on [0, 2pi) following Saltelli et al.
        # Technometrics 1999
        phi = 2 * math.pi * np.random.rand()

        for j in range(D):
            g = 0.5 + (1 / math.pi) * np.arcsin(np.sin(omega2[j] * s + phi))
            X[z, j] = g

    X = scale_samples(X, problem)

    return X


def cli_parse(parser):
    """Add method specific options to CLI parser.

    Parameters
    ----------
    parser : argparse object

    Returns
    -------
    Updated argparse object
    """
    parser.add_argument(
        "-M",
        "--m-coef",
        type=int,
        required=False,
        default=4,
        help="M coefficient, default 4",
        dest="M",
    )

    return parser


def cli_action(args):
    """Run sampling method

    Parameters
    ----------
    args : argparse namespace
    """
    problem = read_param_file(args.paramfile)
    param_values = sample(problem, N=args.samples, M=args.M, seed=args.seed)
    np.savetxt(
        args.output,
        param_values,
        delimiter=args.delimiter,
        fmt="%." + str(args.precision) + "e",
    )


if __name__ == "__main__":
    common_args.run_cli(cli_parse, cli_action)



"""
Created on 30 Jun 2015

@author: will2
"""

import numpy as np
from . import common_args

import pandas as pd
from types import MethodType

from SALib.util import read_param_file, ResultDict
from SALib.sample.ff import generate_contrast, extend_bounds


def analyze(problem, X, Y, second_order=False, print_to_console=False, seed=None):
    """Perform a fractional factorial analysis

    Returns a dictionary with keys 'ME' (main effect) and 'IE' (interaction
    effect). The techniques bulks out the number of parameters with dummy
    parameters to the nearest 2**n.  Any results involving dummy parameters
    could indicate a problem with the model runs.


    Notes
    -----
    Compatible with:
        `ff` : :func:`SALib.sample.ff.sample`


    Examples
    --------
        >>> X = sample(problem)
        >>> Y = X[:, 0] + (0.1 * X[:, 1]) + ((1.2 * X[:, 2]) * (0.2 + X[:, 0]))
        >>> analyze(problem, X, Y, second_order=True, print_to_console=True)


    Parameters
    ----------
    problem: dict
        The problem definition
    X: numpy.matrix
        The NumPy matrix containing the model inputs
    Y: numpy.array
        The NumPy array containing the model outputs
    second_order: bool, default=False
        Include interaction effects
    print_to_console: bool, default=False
        Print results directly to console
    seed : int
        Seed to generate a random number


    Returns
    -------
    Si: dict
        A dictionary of sensitivity indices, including main effects ``ME``,
        and interaction effects ``IE`` (if ``second_order`` is True)


    References
    ----------
    1. Saltelli, A., Ratto, M., Andres, T., Campolongo, F.,
           Cariboni, J., Gatelli, D.,
           Saisana, M., Tarantola, S., 2008.
           Global Sensitivity Analysis: The Primer.
           Wiley, West Sussex, U.K.
           http://doi.org/10.1002/9780470725184
    """
    if seed:
        np.random.seed(seed)

    problem = extend_bounds(problem)
    num_vars = problem["num_vars"]

    X = generate_contrast(problem)

    main_effect = (1.0 / (2 * num_vars)) * np.dot(Y, X)

    Si = ResultDict((k, [None] * num_vars) for k in ["names", "ME"])
    Si["ME"] = main_effect
    Si["names"] = problem["names"]

    if second_order:
        interaction_names, interaction_effects = interactions(problem, Y)
        Si["interaction_names"] = interaction_names
        Si["IE"] = interaction_effects

    Si.to_df = MethodType(to_df, Si)

    if print_to_console:
        for S in Si.to_df():
            print(S)

    return Si


def to_df(self):
    """Conversion method to Pandas DataFrame. To be attached to ResultDict.

    Returns
    -------
    main_effect, inter_effect: tuple
        A tuple of DataFrames for main effects and interaction effects.
        The second element (for interactions) will be `None` if not available.
    """
    names = self["names"]
    main_effect = self["ME"]
    interactions = self.get("IE", None)

    inter_effect = None
    if interactions:
        interaction_names = self.get("interaction_names")
        names = [name for name in names if not isinstance(name, list)]
        inter_effect = pd.DataFrame({"IE": interactions}, index=interaction_names)

    main_effect = pd.DataFrame({"ME": main_effect}, index=names)

    return main_effect, inter_effect


def interactions(problem, Y):
    """Computes the second order effects

    Computes the second order effects (interactions) between
    all combinations of pairs of input factors

    Parameters
    ----------
    problem: dict
        The problem definition
    Y: numpy.array
        The NumPy array containing the model outputs

    Returns
    -------
    ie_names: list
        The names of the interaction pairs
    IE: list
        The sensitivity indices for the pairwise interactions
    """
    names = problem["names"]
    num_vars = problem["num_vars"]

    X = generate_contrast(problem)

    ie_names = []
    IE = []

    for col in range(X.shape[1]):
        for col_2 in range(col):
            x = X[:, col] * X[:, col_2]
            var_names = (names[col_2], names[col])
            ie_names.append(var_names)
            IE.append((1.0 / (2 * num_vars)) * np.dot(Y, x))

    return ie_names, IE


def cli_parse(parser):
    parser.add_argument(
        "-X",
        "--model-input-file",
        type=str,
        required=True,
        default=None,
        help="Model input file",
    )
    parser.add_argument(
        "--max-order",
        type=int,
        required=False,
        default=2,
        choices=[1, 2],
        help="Maximum order of sensitivity \
                           indices to calculate",
    )
    return parser


def cli_action(args):
    problem = read_param_file(args.paramfile)
    Y = np.loadtxt(
        args.model_output_file, delimiter=args.delimiter, usecols=(args.column,)
    )
    X = np.loadtxt(args.model_input_file, delimiter=args.delimiter, ndmin=2)
    if len(X.shape) == 1:
        X = X.reshape((len(X), 1))
    analyze(problem, X, Y, (args.max_order == 2), print_to_console=True, seed=args.seed)


if __name__ == "__main__":
    common_args.run_cli(cli_parse, cli_action)



from typing import Dict
from types import MethodType

import itertools
import time

import numpy as np
from numpy.linalg import solve as lin_solve
from numpy.linalg import svd
from numpy import identity

import pandas as pd
from scipy import stats, special, interpolate

from . import common_args
from ..util import read_param_file, ResultDict

from SALib.plotting.hdmr import plot as hdmr_plot
from SALib.util.problem import ProblemSpec


__all__ = ["analyze", "cli_parse", "cli_action"]


def analyze(
    problem: Dict,
    X: np.ndarray,
    Y: np.ndarray,
    maxorder: int = 2,
    maxiter: int = 100,
    m: int = 2,
    K: int = 20,
    R: int = None,
    alpha: float = 0.95,
    lambdax: float = 0.01,
    print_to_console: bool = False,
    seed: int = None,
) -> Dict:
    """Compute global sensitivity indices using the meta-modeling technique
    known as High-Dimensional Model Representation (HDMR).

    HDMR itself is not a sensitivity analysis method but a surrogate modeling
    approach. It constructs a map of relationship between sets of high
    dimensional inputs and output system variables [1]. This I/O relation can
    be constructed using different basis functions (orthonormal polynomials,
    splines, etc.). The model decomposition can be expressed as

    .. math::
        \\widehat{y} = \\sum_{u \\subseteq \\{1, 2, ..., d \\}} f_u

    where :math:`u` represents any subset including an empty set.

    HDMR becomes extremely useful when the computational cost of obtaining
    sufficient Monte Carlo samples are prohibitive, as may be the case with
    Sobol's method. It uses least-square regression to reduce the required
    number of samples and thus the number of function (model) evaluations.
    Another advantage of this method is that it can account for correlation
    among the model input. Unlike other variance-based methods, the main
    effects are the combination of structural (uncorrelated) and
    correlated contributions.

    This method uses as input

    - a N x d matrix of N different d-vectors of model inputs (factors/parameters)
    - a N x 1 vector of corresponding model outputs

    Notes
    -----
    Compatible with:
        all samplers

    Sets an `emulate` method allowing re-use of the emulator.

    Examples
    --------
    .. code-block:: python
        :linenos:

        sp = ProblemSpec({
            'names': ['X1', 'X2', 'X3'],
            'bounds': [[-np.pi, np.pi]] * 3,
            # 'groups': ['A', 'B', 'A'],
            'outputs': ['Y']
        })

        (sp.sample_saltelli(2048)
            .evaluate(Ishigami.evaluate)
            .analyze_hdmr()
        )

        sp.emulate()

    Parameters
    ----------
    problem : dict
        The problem definition

    X : numpy.matrix
        The NumPy matrix containing the model inputs, N rows by d columns

    Y : numpy.array
        The NumPy array containing the model outputs for each row of X

    maxorder : int (1-3, default: 2)
        Maximum HDMR expansion order

    maxiter : int (1-1000, default: 100)
        Max iterations backfitting

    m : int (2-10, default: 2)
        Number of B-spline intervals

    K : int (1-100, default: 20)
        Number of bootstrap iterations

    R : int (100-N/2, default: N/2)
        Number of bootstrap samples. Will be set to length of `Y` if `K` is set to 1.

    alpha : float (0.5-1)
        Confidence interval F-test

    lambdax : float (0-10, default: 0.01)
        Regularization term

    print_to_console : bool
        Print results directly to console (default: False)

    seed : bool
        Seed to generate a random number

    Returns
    -------
    Si : ResultDict,
        Sa : Uncorrelated contribution of a term

        Sa_conf : Confidence interval of Sa

        Sb : Correlated contribution of a term

        Sb_conf : Confidence interval of Sb

        S : Total contribution of a particular term
            Sum of Sa and Sb, representing first/second/third order sensitivity indices

        S_conf : Confidence interval of S

        ST : Total contribution of a particular dimension/parameter

        ST_conf : Confidence interval of ST

        select : Number of selection (F-Test)

        Em : Emulator result set
            C1: First order coefficient
            C2: Second order coefficient
            C3: Third Order coefficient

    References
    ----------
    1. Rabitz, H. and Aliş, Ö.F.,
       "General foundations of high dimensional model representations",
       Journal of Mathematical Chemistry 25, 197-233 (1999)
       https://doi.org/10.1023/A:1019188517934

    2. Genyuan Li, H. Rabitz, P.E. Yelvington, O.O. Oluwole, F. Bacon,
       C.E. Kolb, and J. Schoendorf,
       "Global Sensitivity Analysis for Systems with Independent and/or
       Correlated Inputs",
       Journal of Physical Chemistry A, Vol. 114 (19), pp. 6022 - 6032, 2010,
       https://doi.org/10.1021/jp9096919
    """
    # Random Seed
    if seed:
        np.random.seed(seed)

    # Initial part: Check input arguments and define HDMR variables
    settings = _check_settings(X, Y, maxorder, maxiter, m, K, R, alpha, lambdax)
    init_vars = _init(X, Y, settings)

    # Sensitivity Analysis Computation with/without bootstraping
    SA, Em, RT, Y_em, idx = _compute(X, Y, settings, init_vars)

    # Finalize results
    Si = _finalize(
        problem, SA, Em, problem["num_vars"], alpha, maxorder, RT, Y_em, idx, X, Y
    )

    # Print results to console
    if print_to_console:
        _print(Si, problem["num_vars"])

    return Si


def _check_settings(X, Y, maxorder, maxiter, m, K, R, alpha, lambdax):
    """Perform checks to ensure all parameters are within usable/expected ranges."""
    # Get dimensionality of numpy arrays
    N, d = X.shape
    y_row = Y.shape[0]

    # Now check input-output mismatch
    if d == 1:
        raise RuntimeError(
            "Matrix X contains only a single column: No point to do"
            " sensitivity analysis when d = 1."
        )
    if N < 300:
        raise RuntimeError(
            f"Number of samples in matrix X, {N}, is insufficient. Need at least 300."
        )
    if N != y_row:
        raise RuntimeError(
            f"Dimension mismatch. The number of outputs ({y_row}) should match"
            f" number of samples ({N})"
        )
    if Y.size != N:
        raise RuntimeError(
            "Y should be a N x 1 vector with one simulated output for each"
            " N parameter vectors."
        )

    if maxorder not in (1, 2, 3):
        raise RuntimeError(
            f'Field "maxorder" of options should be an integer with values of'
            f" 1, 2 or 3, got {maxorder}"
        )

    # Important next check for maxorder - as maxorder relates to d
    if (d == 2) and (maxorder > 2):
        raise RuntimeError(
            'SALib-HDMR ERRROR: Field "maxorder" of options has to be 2 as'
            " d = 2 (X has two columns)"
        )

    if maxiter not in np.arange(1, 1001):
        raise RuntimeError(
            'Field "maxiter" of options should be an integer between 1 to 1000.'
        )

    if m not in np.arange(1, 11):
        raise RuntimeError('Field "m" of options should be an integer between 1 to 10.')

    if K not in np.arange(1, 101):
        raise RuntimeError(
            'Field "K" of options should be an integer between 1 to 100.'
        )

    if R is None:
        R = y_row // 2
    elif R not in np.arange(300, N + 1):
        raise RuntimeError(
            f'Field "R" of options should be an integer between 300 and {N},'
            f" the number of rows matrix X."
        )

    if (K == 1) and (R != y_row):
        R = y_row

    if alpha < 0.5 or alpha > 1.0:
        raise RuntimeError(
            'Field "alpha" of options should be an integer between 0.5 to 1.0'
        )

    if lambdax > 10.0:
        raise RuntimeError(
            'SALib-HDMR WARNING: Field "lambdax" of options set rather large.'
            " Default: lambdax = 0.01"
        )
    elif lambdax < 0.0:
        raise RuntimeError(
            'Field "lambdax" (regularization term) of options cannot be'
            " smaller than zero. Default: 0.01"
        )

    return [N, d, maxorder, maxiter, m, K, R, alpha, lambdax]


def _compute(X, Y, settings, init_vars):
    """Compute HDMR"""
    N, d, maxorder, maxiter, m, K, R, alpha, lambdax = settings
    Em, idx, SA, RT, Y_em, Y_id, m1, m2, m3, j1, j2, j3 = init_vars

    # DYNAMIC PART: Bootstrap for confidence intervals
    for k in range(K):
        tic = time.time()  # Start timer

        # Extract the "right" Y values
        Y_id[:] = Y[idx[:, k]]

        # Compute the variance of the model output
        SA["V_Y"][k, 0] = np.var(Y_id)

        # Mean of the output
        Em["f0"][k] = np.sum(Y_id) / R

        # Compute residuals
        Y_res = Y_id - Em["f0"][k]

        # 1st order component functions: ind/backfitting
        Y_em[:, j1], Y_res, Em["C1"][:, :, k] = _first_order(
            Em["B1"][idx[:, k], :, :],
            Y_res,
            Em["C1"][:, :, k],
            R,
            Em["n1"],
            m1,
            maxiter,
            lambdax,
        )

        # 2nd order component functions: individual
        if maxorder > 1:
            Y_em[:, j2], Y_res, Em["C2"][:, :, k] = _second_order(
                Em["B2"][idx[:, k], :, :],
                Y_res,
                Em["C2"][:, :, k],
                R,
                Em["n2"],
                m2,
                lambdax,
            )

        # 3rd order component functions: individual
        if maxorder == 3:
            Y_em[:, j3], Em["C3"][:, :, k] = _third_order(
                Em["B3"][idx[:, k], :, :],
                Y_res,
                Em["C3"][:, :, k],
                R,
                Em["n3"],
                m3,
                lambdax,
            )

        # Identify significant and insignificant terms
        Em["select"][:, k] = f_test(
            Y_id,
            Em["f0"][k],
            Y_em,
            R,
            alpha,
            m1,
            m2,
            m3,
            Em["n1"],
            Em["n2"],
            Em["n3"],
            Em["n"],
        )

        # Store the emulated output
        Em["Y_e"][:, k] = Em["f0"][k] + np.sum(Y_em, axis=1)

        # RMSE of emulator
        Em["RMSE"][k] = np.sqrt(np.sum(np.square(Y_id - Em["Y_e"][:, k])) / R)

        # Compute sensitivity indices
        SA["S"][:, k], SA["Sa"][:, k], SA["Sb"][:, k] = ancova(
            Y_id, Y_em, SA["V_Y"][k], R, Em["n"]
        )

        RT[k] = time.time() - tic  # Compute CPU time kth emulator

    return (SA, Em, RT, Y_em, idx)


def _init(X, Y, settings):
    """Initialize HDMR variables."""
    N, d, maxorder, maxiter, m, K, R, alpha, lambdax = settings

    # Setup Bootstrap (if K > 1)
    if K == 1:
        # No Bootstrap
        idx = np.arange(0, N).reshape(N, 1)
    else:
        # Now setup the boostrap matrix with selection matrix, idx, for samples
        idx = np.argsort(np.random.rand(N, K), axis=0)[:R]

    # Compute normalized X-values
    X_n = (X - np.tile(X.min(0), (N, 1))) / np.tile((X.max(0)) - X.min(0), (N, 1))

    # DICT Em: Important variables
    n2 = 0
    n3 = 0
    c2 = np.empty
    c3 = np.empty

    # determine all combinations of parameters (coefficients) for each order
    c1 = np.arange(0, d)
    n1 = d

    # now return based on maxorder used
    if maxorder > 1:
        c2 = np.asarray(list(itertools.combinations(np.arange(0, d), 2)))
        n2 = c2.shape[0]
    if maxorder == 3:
        c3 = np.asarray(list(itertools.combinations(np.arange(0, d), 3)))
        n3 = c3.shape[0]

    # calulate total number of coefficients
    n = n1 + n2 + n3

    # Initialize m1, m2 and m3 - number of coefficients first, second, third
    # order
    m1 = m + 3
    m2 = m1**2
    m3 = m1**3

    # STRUCTURE Em: Initialization
    Em = {
        "nterms": np.zeros(K),
        "RMSE": np.zeros(K),
        "m": m,
        "Y_e": np.zeros((R, K)),
        "f0": np.zeros(K),
        "c1": c1,
        "C1": np.zeros((m1, n1, K)),
        "C2": np.zeros((1, 1, K)),
        "C3": np.zeros((1, 1, K)),
        "n": n,
        "n1": n1,
        "n2": n2,
        "n3": n3,
        "maxorder": maxorder,
        "select": np.zeros((n, K)),
        "B1": np.zeros((N, m1, n1)),
    }

    if maxorder >= 2:
        Em.update({"c2": c2, "B2": np.zeros((N, m2, n2)), "C2": np.zeros((m2, n2, K))})
    if maxorder >= 3:
        Em.update({"c3": c3, "B3": np.zeros((N, m3, n3)), "C3": np.zeros((m3, n3, K))})

    # Compute cubic B-spline
    Em["B1"] = B_spline(X_n, m, d)

    # Now compute B values for second order
    if maxorder > 1:
        beta = np.array(list(itertools.product(range(m1), repeat=2)))
        for k, j in itertools.product(range(n2), range(m2)):
            Em["B2"][:, j, k] = np.multiply(
                Em["B1"][:, beta[j, 0], Em["c2"][k, 0]],
                Em["B1"][:, beta[j, 1], Em["c2"][k, 1]],
            )

    # Compute B values for third order
    if maxorder == 3:
        # Now compute B values for third order
        beta = np.array(list(itertools.product(range(m1), repeat=3)))
        EmB1 = Em["B1"]

        for k, j in itertools.product(range(n3), range(m3)):
            Emc3_k = Em["c3"][k]
            beta_j = beta[j]
            Em["B3"][:, j, k] = np.multiply(
                np.multiply(
                    EmB1[:, beta_j[0], Emc3_k[0]], EmB1[:, beta_j[1], Emc3_k[1]]
                ),
                EmB1[:, beta_j[2], Emc3_k[2]],
            )

    # DICT SA: Sensitivity analysis and analysis of variance decomposition
    em_k = np.zeros((Em["n"], K))
    SA = {
        "S": em_k,
        "Sa": em_k.copy(),
        "Sb": em_k.copy(),
        "ST": np.zeros((d, K)),
        "V_Y": np.zeros((K, 1)),
    }

    # Return runtime
    RT = np.zeros(K)

    # Initialize emulator matrix
    Y_em = np.zeros((R, Em["n"]))

    # Initialize index of first, second and third order terms (columns of Y_em)
    j1 = range(n1)
    j2 = range(n1, n1 + n2)
    j3 = range(n1 + n2, n1 + n2 + n3)

    # Initialize temporary Y_id for bootstrap
    Y_id = np.zeros(R)

    return [Em, idx, SA, RT, Y_em, Y_id, m1, m2, m3, j1, j2, j3]


def B_spline(X, m, d):
    """Generate cubic B-splines using scipy basis_element method.

    Knot points (`t`) are automatically determined.

    References
    ----------
    .. [1] https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.BSpline.basis_element.html  # noqa: E501
    """
    # Initialize B matrix
    B = np.zeros((X.shape[0], m + 3, d))

    # Cubic basis-spline settings
    k = np.arange(-1, m + 2)

    # Compute B-Splines
    for j, i in itertools.product(range(d), range(m + 3)):
        k_i = k[i]
        t = np.arange(k_i - 2, k_i + 3) / m
        temp = interpolate.BSpline.basis_element(t)(X[:, j]) * np.power(m, 3)
        B[:, i, j] = np.where(temp < 0, 0, temp)

    return B


def _first_order(B1, Y_res, C1, R, n1, m1, maxiter, lambdax):
    """Compute first order sensitivities."""
    Y_i = np.empty((R, n1))  # Initialize 1st order contributions
    T1 = np.empty((m1, R, n1))  # Initialize T(emporary) matrix - 1st
    it = 0  # Initialize iteration counter

    lam_eye_m1 = lambdax * identity(m1)

    # Avoid memory allocations by using temporary store
    B1_j = np.empty(B1[:, :, 0].shape)
    B11 = np.empty((m1, m1))

    # First order individual estimation
    for j in range(n1):
        # Regularized least squares inversion ( minimize || C1 ||_2 )
        B1_j[:, :] = B1[:, :, j]
        B11[:, :] = B1_j.T @ B1_j

        # if it is ill-conditioned matrix, the default value is zero
        if np.all(svd(B11)[1]):  # sigma, diagonal matrix, from svd
            T1[:, :, j] = lin_solve(B11 + lam_eye_m1, B1_j.T)

        C1[:, j] = T1[:, :, j] @ Y_res
        Y_i[:, j] = B1_j @ C1[:, j]

    # Backfitting Method
    var1b_old = np.sum(np.square(C1), axis=0)
    varmax = 1
    while (varmax > 1e-3) and (it < maxiter):
        for j in range(n1):
            Y_r = Y_res
            for z in range(n1):
                if j != z:
                    Y_r = Y_r - B1[:, :, z] @ C1[:, z]

            C1[:, j] = T1[:, :, j] @ Y_r

        var1b_new = np.sum(np.square(C1), axis=0)
        varmax = np.max(np.absolute(var1b_new - var1b_old))
        var1b_old = var1b_new
        it += 1

    # Now compute first-order terms
    for j in range(n1):
        Y_i[:, j] = B1[:, :, j] @ C1[:, j]

        # Subtract each first order term from residuals
        Y_res = Y_res - Y_i[:, j]

    return (Y_i, Y_res, C1)


def _second_order(B2, Y_res, C2, R, n2, m2, lambdax):
    """Compute second order sensitivities."""
    Y_ij = np.empty((R, n2))  # Initialize 2nd order contributions
    T2 = np.empty((m2, R))  # Initialize T(emporary) matrix - 1st

    lam_eye_m2 = lambdax * identity(m2)

    # Avoid memory allocations by using temporary store
    B2_j = np.empty(B2[:, :, 0].shape)
    B22 = np.empty((m2, m2))

    # First order individual estimation
    for j in range(n2):
        # Regularized least squares inversion ( minimize || C1 ||_2 )
        B2_j[:, :] = B2[:, :, j]
        B22[:, :] = B2_j.T @ B2_j

        # if it is ill-conditioned matrix, the default value is zero
        if np.all(svd(B22)[1]):  # sigma, diagonal matrix, from svd
            T2[:, :] = lin_solve(B22 + lam_eye_m2, B2_j.T)
        else:
            T2[:, :] = 0.0

        C2[:, j] = T2 @ Y_res
        Y_ij[:, j] = B2_j @ C2[:, j]

    # Now compute second-order terms
    for j in range(n2):
        # Subtract each first order term from residuals
        Y_res = Y_res - Y_ij[:, j]

    return (Y_ij, Y_res, C2)


def _third_order(B3, Y_res, C3, R, n3, m3, lambdax):
    """Compute third order sensitivities."""
    Y_ijk = np.empty((R, n3))  # Initialize 3rd order contributions
    T3 = np.empty((m3, R))  # Initialize T(emporary) matrix - 1st

    lam_eye_m3 = lambdax * identity(m3)

    # Avoid memory allocations by using temporary store
    B3_j = np.empty(B3[:, :, 0].shape)
    B33 = np.empty((m3, m3))

    # First order individual estimation
    for j in range(n3):
        # Regularized least squares inversion ( minimize || C1 ||_2 )
        B3_j[:, :] = B3[:, :, j]
        B33[:, :] = B3_j.T @ B3_j

        # if it is ill-conditioned matrix, the default value is zero
        if np.all(svd(B33)[1]):  # sigma, diagonal matrix, from svd
            T3[:, :] = lin_solve(B33 + lam_eye_m3, B3_j.T)
        else:
            T3[:, :] = 0.0

        C3[:, j] = T3 @ Y_res
        Y_ijk[:, j] = B3_j @ C3[:, j]

    return (Y_ijk, C3)


def f_test(Y, f0, Y_em, R, alpha, m1, m2, m3, n1, n2, n3, n):
    """Use F-test for model selection."""
    # Initialize ind with zeros (all terms insignificant)
    select = np.zeros(n)

    # Determine the significant components of the HDMR model via the F-test
    Y_res0 = Y - f0
    SSR0 = np.sum(np.square(Y_res0))
    p0 = 0
    for i in range(n):
        # model with ith term included
        Y_res1 = Y_res0 - Y_em[:, i]

        # Number of parameters of proposed model (order dependent)
        if i <= n1:
            p1 = m1  # 1st order
        elif n1 < i <= (n1 + n2):
            p1 = m2  # 2nd order
        else:
            p1 = m3  # 3rd order

        # Calculate SSR of Y1
        SSR1 = np.sum(np.square(Y_res1))

        # Now calculate the F_stat (F_stat > 0 -> SSR1 < SSR0 )
        F_stat = ((SSR0 - SSR1) / (p1 - p0)) / (SSR1 / (R - p1))

        # Now calculate critical F value
        F_crit = stats.f.ppf(q=alpha, dfn=p1 - p0, dfd=R - p1)

        # Now determine whether to accept ith component into model
        if F_stat > F_crit:
            # ith term is significant and should be included in model
            select[i] = 1

    return select


def ancova(Y, Y_em, V_Y, R, n):
    """Analysis of Covariance."""
    # Compute the sum of all Y_em terms
    Y0 = np.sum(Y_em, axis=1)

    # Initialize each variable
    S, S_a, S_b = np.zeros((3, n))

    # Analysis of covariance
    for j in range(n):
        # Covariance matrix of jth term of Y_em and actual Y
        C = np.cov(np.stack((Y_em[:, j], Y), axis=0))

        # Total sensitivity of jth term         ( = Eq. 19 of Li et al )
        S[j] = C[0, 1] / V_Y

        # Covariance matrix of jth term with emulator Y without jth term
        C = np.cov(np.stack((Y_em[:, j], Y0 - Y_em[:, j]), axis=0))

        # Structural contribution of jth term   ( = Eq. 20 of Li et al )
        S_a[j] = C[0, 0] / V_Y

        # Correlative contribution of jth term  ( = Eq. 21 of Li et al )
        S_b[j] = C[0, 1] / V_Y

    return (S, S_a, S_b)


def _finalize(problem, SA, Em, d, alpha, maxorder, RT, Y_em, bootstrap_idx, X, Y):
    """Creates ResultDict to be returned."""

    # Create Sensitivity Indices Result Dictionary
    keys = (
        "Sa",
        "Sa_conf",
        "Sb",
        "Sb_conf",
        "S",
        "S_conf",
        "select",
        "Sa_sum",
        "Sa_sum_conf",
        "Sb_sum",
        "Sb_sum_conf",
        "S_sum",
        "S_sum_conf",
    )
    Si = ResultDict((k, np.zeros(Em["n"])) for k in keys)
    Si["Term"] = [None] * Em["n"]
    Si["ST"] = [
        np.nan,
    ] * Em["n"]
    Si["ST_conf"] = [
        np.nan,
    ] * Em["n"]

    # Z score
    def z(p):
        return (-1) * np.sqrt(2) * special.erfcinv(p * 2)

    # Multiplier for confidence interval
    mult = z(alpha + (1 - alpha) / 2)

    # Compute the total sensitivity of each parameter/coefficient
    for r in range(Em["n1"]):
        if maxorder == 1:
            TS = SA["S"][r, :]
        elif maxorder == 2:
            ij = Em["n1"] + np.where(np.sum(Em["c2"] == r, axis=1) == 1)[0]
            TS = np.sum(SA["S"][np.append(r, ij), :], axis=0)
        elif maxorder == 3:
            ij = Em["n1"] + np.where(np.sum(Em["c2"] == r, axis=1) == 1)[0]
            ijk = Em["n1"] + Em["n2"] + np.where(np.sum(Em["c3"] == r, axis=1) == 1)[0]
            TS = np.sum(SA["S"][np.append(r, np.append(ij, ijk)), :], axis=0)
        Si["ST"][r] = np.mean(TS)
        Si["ST_conf"][r] = mult * np.std(TS)

    # Fill Expansion Terms
    ct = 0
    p_names = problem["names"]
    p_c2 = Em["c2"]
    # p_c3 = Em['c3']
    for i in range(Em["n1"]):
        Si["Term"][ct] = p_names[i]
        ct += 1

    for i in range(Em["n2"]):
        Si["Term"][ct] = "/".join([p_names[p_c2[i, 0]], p_names[p_c2[i, 1]]])
        ct += 1

    for i in range(Em["n3"]):
        Si["Term"][ct] = "/".join(
            [p_names[Em["c3"][i, 0]], p_names[Em["c3"][i, 1]], p_names[Em["c3"][i, 2]]]
        )
        ct += 1

    # Assign Bootstrap Results to Si Dict
    Si["Sa"] = np.mean(SA["Sa"], axis=1)
    Si["Sb"] = np.mean(SA["Sb"], axis=1)
    Si["S"] = np.mean(SA["S"], axis=1)
    Si["Sa_sum"] = np.mean(np.sum(SA["Sa"], axis=0))
    Si["Sb_sum"] = np.mean(np.sum(SA["Sb"], axis=0))
    Si["S_sum"] = np.mean(np.sum(SA["S"], axis=0))

    # Compute Confidence Interval
    Si["Sa_conf"] = mult * np.std(SA["Sa"], axis=1)
    Si["Sb_conf"] = mult * np.std(SA["Sb"], axis=1)
    Si["S_conf"] = mult * np.std(SA["S"], axis=1)
    Si["Sa_sum_conf"] = mult * np.std(np.sum(SA["Sa"]))
    Si["Sb_sum_conf"] = mult * np.std(np.sum(SA["Sb"]))
    Si["S_sum_conf"] = mult * np.std(np.sum(SA["S"]))

    # F-test # of selection to print out
    Si["select"] = Em["select"].flatten()

    Si["names"] = Si["Term"]

    # Additional data for plotting.
    Si["Em"] = Em
    Si["RT"] = RT
    Si["Y_em"] = Y_em
    Si["idx"] = bootstrap_idx
    Si["X"] = X
    Si["Y"] = Y

    Si.emulate = MethodType(emulate, Si)
    Si.to_df = MethodType(to_df, Si)

    Si._plot = Si.plot
    Si.plot = MethodType(plot, Si)

    if not isinstance(problem, ProblemSpec):
        Si.problem = problem
    else:
        problem.update(Si)
        problem.emulate = MethodType(emulate, problem)

    return Si


def emulate(self, X, Y=None):
    """Emulates model output with new input data.

    Generates B-Splines with new input matrix, X, and multiplies it with
    coefficient matrix, C.

    Compares emulated results with observed vector, Y.

    Returns
    ========
    Si : ResultDict, Updated

    """
    # Dimensionality
    N, d = X.shape

    # Expand C
    Em = self["Em"]
    C1, C2, C3 = Em["C1"], Em["C2"], Em["C3"]

    # Take average coefficient
    C1 = np.mean(C1, axis=2)
    C2 = np.mean(C2, axis=2)
    C3 = np.mean(C3, axis=2)

    # Get maxorder and other information from C matrix
    m = C1.shape[0] - 3
    m1 = m + 3
    n1, n2, n3 = d, 0, 0
    if C2.shape[0] != 1:
        maxorder = 2
        m2 = C2.shape[0]
        c2 = np.asarray(list(itertools.combinations(np.arange(0, d), 2)))
        n2 = c2.shape[0]
    if C3.shape[0] != 1:
        maxorder = 3
        m3 = C3.shape[0]
        c3 = np.asarray(list(itertools.combinations(np.arange(0, d), 3)))
        n3 = c3.shape[0]

    # Initialize emulated Y's
    Y_em = np.zeros((N, n1 + n2 + n3))

    # Compute normalized X-values
    X_n = (X - np.tile(X.min(0), (N, 1))) / np.tile((X.max(0)) - X.min(0), (N, 1))

    # Now compute B values
    B1 = B_spline(X_n, m, d)
    if maxorder > 1:
        B2 = np.zeros((N, m2, n2))
        beta = np.array(list(itertools.product(range(m1), repeat=2)))
        for k, j in itertools.product(range(n2), range(m2)):
            B2[:, j, k] = np.multiply(
                B1[:, beta[j, 0], c2[k, 0]], B1[:, beta[j, 1], c2[k, 1]]
            )
    if maxorder == 3:
        B3 = np.zeros((N, m3, n3))
        beta = np.array(list(itertools.product(range(m1), repeat=3)))
        for k, j in itertools.product(range(n3), range(m3)):
            Emc3_k = c3[k]
            beta_j = beta[j]
            B3[:, j, k] = np.multiply(
                np.multiply(B1[:, beta_j[0], Emc3_k[0]], B1[:, beta_j[1], Emc3_k[1]]),
                B1[:, beta_j[2], Emc3_k[2]],
            )

    # Calculate emulated output: First Order
    for j in range(0, n1):
        Y_em[:, j] = B1[:, :, j] @ C1[:, j]

    # Second Order
    if maxorder > 1:
        for j in range(n1, n1 + n2):
            Y_em[:, j] = B2[:, :, j - n1] @ C2[:, j - n1]

    # Third Order
    if maxorder == 3:
        for j in range(n1 + n2, n1 + n2 + n3):
            Y_em[:, j] = B3[:, :, j - n1 - n2] @ C3[:, j - n1 - n2]

    Y_mean = 0
    if Y is not None:
        Y_mean = np.mean(Y)
        self["Y_test"] = Y

    self["emulated"] = np.sum(Y_em, axis=1) + Y_mean


def to_df(self):
    """Conversion method to Pandas DataFrame. To be attached to ResultDict.

    Returns
    -------
    Pandas DataFrame
    """
    names = self["names"]

    # Only convert these elements in dict to DF
    include_list = ["Sa", "Sb", "S", "ST"]
    include_list += [f"{name}_conf" for name in include_list]
    new_spec = {k: v for k, v in self.items() if k in include_list}

    return pd.DataFrame(new_spec, index=names)


def plot(self):
    """HDMR specific plotting.

    Gets attached to the SALib problem spec.
    """
    return hdmr_plot(self)


def _print(Si, d):
    print("\n")
    print(
        "Term    \t      Sa            Sb             S             ST         #select "
    )
    print("-" * 84)  # Header break

    format1 = "%-11s   \t %5.2f (\261%.2f) %5.2f (\261%.2f) %5.2f (\261%.2f) %5.2f (\261%.2f)    %-3.0f"  # noqa: E501
    format2 = "%-11s   \t %5.2f (\261%.2f) %5.2f (\261%.2f) %5.2f (\261%.2f)                  %-3.0f"  # noqa: E501

    for i in range(len(Si["Sa"])):
        if i < d:
            print(
                format1
                % (
                    Si["Term"][i],
                    Si["Sa"][i],
                    Si["Sa_conf"][i],
                    Si["Sb"][i],
                    Si["Sb_conf"][i],
                    Si["S"][i],
                    Si["S_conf"][i],
                    Si["ST"][i],
                    Si["ST_conf"][i],
                    np.sum(Si["select"][i]),
                )
            )
        else:
            print(
                format2
                % (
                    Si["Term"][i],
                    Si["Sa"][i],
                    Si["Sa_conf"][i],
                    Si["Sb"][i],
                    Si["Sb_conf"][i],
                    Si["S"][i],
                    Si["S_conf"][i],
                    np.sum(Si["select"][i]),
                )
            )

    print("-" * 84)  # Header break

    format3 = "%-11s   \t %5.2f (\261%.2f) %5.2f (\261%.2f) %5.2f (\261%.2f)"
    print(
        format3
        % (
            "Sum",
            Si["Sa_sum"],
            Si["Sa_sum_conf"],
            Si["Sb_sum"],
            Si["Sb_sum_conf"],
            Si["S_sum"],
            Si["S_sum_conf"],
        )
    )

    keys = ("Sa_sum", "Sb_sum", "S_sum", "Sa_sum_conf", "Sb_sum_conf", "S_sum_conf")
    for k in keys:
        Si.pop(k, None)

    return Si


def cli_parse(parser):
    parser.add_argument(
        "-X",
        "--model-input-file",
        type=str,
        required=True,
        default=None,
        help="Model input file",
    )
    parser.add_argument(
        "-mor",
        "--maxorder",
        type=int,
        required=False,
        default=2,
        help="Maximum order of expansion 1-3",
    )
    parser.add_argument(
        "-mit",
        "--maxiter",
        type=int,
        required=False,
        default=100,
        help="Maximum iteration number",
    )
    parser.add_argument(
        "-m", "--m-int", type=int, required=False, default=2, help="B-spline interval"
    )
    parser.add_argument(
        "-K",
        "--K-bootstrap",
        type=int,
        required=False,
        default=20,
        help="Number of bootstrap",
    )
    parser.add_argument(
        "-R",
        "--R-subsample",
        type=int,
        required=False,
        default=None,
        help="Subsample size",
    )
    parser.add_argument(
        "-a",
        "--alpha",
        type=float,
        required=False,
        default=0.95,
        help="Confidence interval",
    )
    parser.add_argument(
        "-lambda",
        "--lambdax",
        type=float,
        required=False,
        default=0.01,
        help="Regularization constant",
    )
    return parser


def cli_action(args):
    problem = read_param_file(args.paramfile)
    Y = np.loadtxt(
        args.model_output_file, delimiter=args.delimiter, usecols=(args.column,)
    )
    X = np.loadtxt(args.model_input_file, delimiter=args.delimiter, ndmin=2)
    mor = args.maxorder
    mit = args.maxiter
    m = args.m_int
    K = args.K_bootstrap
    R = args.R_subsample
    alpha = args.alpha
    lambdax = args.lambdax
    options = options = {
        "maxorder": mor,
        "maxiter": mit,
        "m": m,
        "K": K,
        "R": R,
        "alpha": alpha,
        "lambdax": lambdax,
        "print_to_console": True,
    }

    if len(X.shape) == 1:
        X = X.reshape((len(X), 1))

    analyze(problem, X, Y, **options)


if __name__ == "__main__":
    common_args.run_cli(cli_parse, cli_action)



"""The sampling implementation of fractional factorial method

This implementation is based on the formulation put forward in
[`Saltelli et al. 2008 <http://doi.org/10.1002/9780470725184>`_]
"""

from scipy.linalg import hadamard
import numpy as np
from . import common_args
from ..util import scale_samples, read_param_file


def find_smallest(num_vars):
    """Find the smallest exponent of two that is greater than the number
    of variables.

    Parameters
    ----------
    num_vars : int
        Number of variables

    Returns
    -------
    x : int
        Smallest exponent of two greater than `num_vars`
    """
    for x in range(10):
        if num_vars <= 2**x:
            return x


def extend_bounds(problem):
    """Extends the problem bounds to the nearest power of two.

    Parameters
    ----------
    problem : dict
        The problem definition
    """

    num_vars = problem["num_vars"]
    num_ff_vars = 2 ** find_smallest(num_vars)
    num_dummy_variables = num_ff_vars - num_vars

    bounds = list(problem["bounds"])
    names = problem["names"]
    if num_dummy_variables > 0:
        bounds.extend([[0, 1] for x in range(num_dummy_variables)])
        names.extend(["dummy_" + str(var) for var in range(num_dummy_variables)])
        problem["bounds"] = bounds
        problem["names"] = names
        problem["num_vars"] = num_ff_vars

    return problem


def generate_contrast(problem):
    """Generates the raw sample from the problem file.

    Parameters
    ----------
    problem : dict
        The problem definition
    """

    num_vars = problem["num_vars"]

    # Find the smallest n, such that num_vars < k
    k_chosen = 2 ** find_smallest(num_vars)

    # Generate the fractional factorial contrast
    contrast = np.vstack([hadamard(k_chosen), -hadamard(k_chosen)])

    return contrast


def sample(problem, seed=None):
    """Generates model inputs using a fractional factorial sample.

    Returns a NumPy matrix containing the model inputs required for a
    fractional factorial analysis.
    The resulting matrix has D columns, where D is smallest power of 2 that is
    greater than the number of parameters.
    These model inputs are intended to be used with
    :func:`SALib.analyze.ff.analyze`.

    The problem file is padded with a number of dummy variables called
    ``dummy_0`` required for this procedure. These dummy variables can be used
    as a check for errors in the analyze procedure.

    This algorithm is an implementation of that contained in Saltelli et al
    [`Saltelli et al. 2008 <http://doi.org/10.1002/9780470725184>`_]

    Parameters
    ----------
    problem : dict
        The problem definition
    seed : int
        Seed to generate a random number

    Returns
    -------
    sample : :class:`numpy.array`

    References
    ----------
    1. Saltelli, A., Ratto, M., Andres, T., Campolongo, F.,
       Cariboni, J., Gatelli, D., Saisana, M., Tarantola, S., 2008.
       Global Sensitivity Analysis: The Primer.
       Wiley, West Sussex, U.K.
       http://doi.org/10.1002/9780470725184
    """
    if seed:
        np.random.seed(seed)
    contrast = generate_contrast(problem)
    sample = np.array((contrast + 1.0) / 2, dtype=float)
    problem = extend_bounds(problem)

    sample = scale_samples(sample, problem)
    return sample


# No additional CLI options
cli_parse = None


def cli_action(args):
    """Run sampling method

    Parameters
    ----------
    args : argparse namespace
    """
    problem = read_param_file(args.paramfile)
    param_values = sample(problem, seed=args.seed)
    np.savetxt(
        args.output,
        param_values,
        delimiter=args.delimiter,
        fmt="%." + str(args.precision) + "e",
    )


if __name__ == "__main__":
    common_args.run_cli(cli_parse, cli_action)



import math
import sys

import numpy as np

from .directions import directions


if sys.version_info[0] > 2:
    long = int

# ==============================================================================
# The following code is based on the Sobol sequence generator by Frances
# Y. Kuo and Stephen Joe. The license terms are provided below.
#
# Copyright (c) 2008, Frances Y. Kuo and Stephen Joe
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# Neither the names of the copyright holders nor the names of the
# University of New South Wales and the University of Waikato
# and its contributors may be used to endorse or promote products derived
# from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ==============================================================================


def sample(N, D):
    """Generate (N x D) numpy array of Sobol sequence samples"""
    scale = 31
    result = np.zeros([N, D])

    if D > len(directions) + 1:
        raise ValueError("Error in Sobol sequence: not enough dimensions")

    L = int(math.ceil(math.log(N) / math.log(2)))

    if L > scale:
        raise ValueError("Error in Sobol sequence: not enough bits")

    for i in range(D):
        V = np.zeros(L + 1, dtype=long)

        if i == 0:
            for j in range(1, L + 1):
                V[j] = 1 << (scale - j)  # all m's = 1
        else:
            m = np.array(directions[i - 1], dtype=int)
            a = m[0]
            s = len(m) - 1

            # The following code discards the first row of the ``m`` array
            # Because it has floating point errors, e.g. values of 2.24e-314
            if L <= s:
                for j in range(1, L + 1):
                    V[j] = m[j] << (scale - j)
            else:
                for j in range(1, s + 1):
                    V[j] = m[j] << (scale - j)
                for j in range(s + 1, L + 1):
                    V[j] = V[j - s] ^ (V[j - s] >> s)
                    for k in range(1, s):
                        V[j] ^= ((a >> (s - 1 - k)) & 1) * V[j - k]

        X = long(0)
        for j in range(1, N):
            X ^= V[index_of_least_significant_zero_bit(j - 1)]
            result[j][i] = float(X / math.pow(2, scale))

    return result


def index_of_least_significant_zero_bit(value):
    index = 1
    while (value & 1) != 0:
        value >>= 1
        index += 1

    return index



import numpy as np

from . import common_args
from ..util import read_param_file, scale_samples, compute_groups_matrix


def sample(problem, N, seed=None):
    """Generate model inputs using Latin hypercube sampling (LHS).

    Returns a NumPy matrix containing the model inputs generated by Latin
    hypercube sampling.  The resulting matrix contains N rows and D columns,
    where D is the number of parameters.

    Parameters
    ----------
    problem : dict
        The problem definition
    N : int
        The number of samples to generate
    seed : int
        Seed to generate a random number

    References
    ----------
    1. McKay, M.D., Beckman, R.J., Conover, W.J., 1979.
           A comparison of three methods for selecting values of input
           variables in the analysis of output from a computer code.
           Technometrics 21, 239-245.
           https://doi.org/10.2307/1268522

    2. Iman, R.L., Helton, J.C., Campbell, J.E., 1981.
           An Approach to Sensitivity Analysis of Computer Models:
           Part I—Introduction, Input Variable Selection and
           Preliminary Variable Assessment.
           Journal of Quality Technology 13, 174-183.
           https://doi.org/10.1080/00224065.1981.11978748

    """
    num_samples = N

    if seed:
        np.random.seed(seed)

    groups = problem.get("groups")
    if groups:
        num_groups = len(set(groups))
        G, group_names = compute_groups_matrix(groups)
    else:
        num_groups = problem["num_vars"]

    result = np.empty([num_samples, problem["num_vars"]])
    temp = np.empty([num_samples])
    d = 1.0 / num_samples

    temp = np.array(
        [
            np.random.uniform(low=sample * d, high=(sample + 1) * d, size=num_groups)
            for sample in range(num_samples)
        ]
    )

    for group in range(num_groups):

        np.random.shuffle(temp[:, group])

        for sample in range(num_samples):
            if groups:
                grouped_variables = np.where(G[:, group] == 1)
                result[sample, grouped_variables[0]] = temp[sample, group]
            else:
                result[sample, group] = temp[sample, group]

    result = scale_samples(result, problem)

    return result


# No additional CLI options
cli_parse = None


def cli_action(args):
    """Run sampling method

    Parameters
    ----------
    args : argparse namespace
    """
    problem = read_param_file(args.paramfile)
    param_values = sample(problem, args.samples, seed=args.seed)
    np.savetxt(
        args.output,
        param_values,
        delimiter=args.delimiter,
        fmt="%." + str(args.precision) + "e",
    )


if __name__ == "__main__":
    common_args.run_cli(cli_parse, cli_action)



from typing import Dict
from scipy.stats import norm, gaussian_kde, rankdata

import numpy as np

from . import common_args
from ..util import read_param_file, ResultDict


def analyze(
    problem: Dict,
    X: np.ndarray,
    Y: np.ndarray,
    num_resamples: int = 100,
    conf_level: float = 0.95,
    print_to_console: bool = False,
    seed: int = None,
) -> Dict:
    """Perform Delta Moment-Independent Analysis on model outputs.

    Returns a dictionary with keys 'delta', 'delta_conf', 'S1', and 'S1_conf',
    where each entry is a list of size D (the number of parameters) containing
    the indices in the same order as the parameter file.


    Notes
    -----
    Compatible with:
        all samplers


    Examples
    --------
        >>> X = latin.sample(problem, 1000)
        >>> Y = Ishigami.evaluate(X)
        >>> Si = delta.analyze(problem, X, Y, print_to_console=True)


    Parameters
    ----------
    problem : dict
        The problem definition
    X: numpy.matrix
        A NumPy matrix containing the model inputs
    Y : numpy.array
        A NumPy array containing the model outputs
    num_resamples : int
        The number of resamples when computing confidence intervals (default 100)
    conf_level : float
        The confidence interval level (default 0.95)
    print_to_console : bool
        Print results directly to console (default False)


    References
    ----------
    1. Borgonovo, E. (2007). "A new uncertainty importance measure."
           Reliability Engineering & System Safety, 92(6):771-784,
           doi:10.1016/j.ress.2006.04.015.

    2. Plischke, E., E. Borgonovo, and C. L. Smith (2013). "Global
           sensitivity measures from given data." European Journal of
           Operational Research, 226(3):536-550, doi:10.1016/j.ejor.2012.11.047.
    """
    if seed:
        np.random.seed(seed)

    D = problem["num_vars"]
    N = Y.size

    if not 0 < conf_level < 1:
        raise RuntimeError("Confidence level must be between 0-1.")

    # equal frequency partition
    exp = 2.0 / (7.0 + np.tanh((1500.0 - N) / 500.0))
    M = int(np.round(min(int(np.ceil(N**exp)), 48)))
    m = np.linspace(0, N, M + 1)
    Ygrid = np.linspace(np.min(Y), np.max(Y), 100)

    keys = ("delta", "delta_conf", "S1", "S1_conf")
    S = ResultDict((k, np.zeros(D)) for k in keys)
    S["names"] = problem["names"]

    try:
        for i in range(D):
            X_i = X[:, i]
            S["delta"][i], S["delta_conf"][i] = bias_reduced_delta(
                Y, Ygrid, X_i, m, num_resamples, conf_level
            )
            S["S1"][i] = sobol_first(Y, X_i, m)
            S["S1_conf"][i] = sobol_first_conf(Y, X_i, m, num_resamples, conf_level)
    except np.linalg.LinAlgError as e:
        msg = "Singular matrix detected\n"
        msg += "This may be due to the sample size ({}) being too small\n".format(
            Y.size
        )
        msg += "If this is not the case, check Y values or raise an issue with the\n"
        msg += "SALib team"

        raise np.linalg.LinAlgError(msg) from e

    if print_to_console:
        print(S.to_df())

    return S


def calc_delta(Y, Ygrid, X, m):
    """Plischke et al. (2013) delta index estimator (eqn 26) for d_hat."""
    N = len(Y)
    fy = gaussian_kde(Y, bw_method="silverman")(Ygrid)
    xr = rankdata(X, method="ordinal")

    d_hat = 0.0
    l_m = len(m) - 1
    for j in range(l_m):
        ix = np.where((xr > m[j]) & (xr <= m[j + 1]))[0]
        nm = len(ix)

        # if not np.all(np.equal(Y_ix, Y_ix[0])):
        Y_ix = Y[ix]
        if Y_ix.ptp() != 0.0:
            fyc = gaussian_kde(Y_ix, bw_method="silverman")(Ygrid)
            fy_ = np.abs(fy - fyc)
        else:
            fy_ = np.abs(fy)

        d_hat += (nm / (2 * N)) * np.trapz(fy_, Ygrid)

    return d_hat


def bias_reduced_delta(Y, Ygrid, X, m, num_resamples, conf_level):
    """Plischke et al. 2013 bias reduction technique (eqn 30)"""
    d = np.empty(num_resamples)
    d_hat = calc_delta(Y, Ygrid, X, m)

    N = len(Y)
    r = np.random.randint(N, size=(num_resamples, N))
    for i in range(num_resamples):
        r_i = r[i, :]
        d[i] = calc_delta(Y[r_i], Ygrid, X[r_i], m)

    d = 2.0 * d_hat - d
    return (d.mean(), norm.ppf(0.5 + conf_level / 2) * d.std(ddof=1))


def sobol_first(Y, X, m):
    # pre-process to catch constant array
    # see: https://github.com/numpy/numpy/issues/9631
    if Y.ptp() == 0.0:
        # Catch constant results
        # If Y does not change then it is not sensitive to anything...
        return 0.0

    xr = rankdata(X, method="ordinal")
    Vi = 0
    N = len(Y)
    Y_mean = Y.mean()
    for j in range(len(m) - 1):
        ix = np.where((xr > m[j]) & (xr <= m[j + 1]))[0]
        nm = len(ix)
        Vi += (nm / N) * ((Y[ix].mean() - Y_mean) ** 2)

    return Vi / np.var(Y)


def sobol_first_conf(Y, X, m, num_resamples, conf_level):
    s = np.zeros(num_resamples)

    N = len(Y)
    r = np.random.randint(N, size=(num_resamples, N))
    for i in range(num_resamples):
        r_i = r[i, :]
        s[i] = sobol_first(Y[r_i], X[r_i], m)

    return norm.ppf(0.5 + conf_level / 2) * s.std(ddof=1)


def cli_parse(parser):
    parser.add_argument(
        "-X",
        "--model-input-file",
        type=str,
        required=True,
        default=None,
        help="Model input file",
    )
    parser.add_argument(
        "-r",
        "--resamples",
        type=int,
        required=False,
        default=10,
        help="Number of bootstrap resamples for \
                           Sobol confidence intervals",
    )
    return parser


def cli_action(args):
    problem = read_param_file(args.paramfile)
    Y = np.loadtxt(
        args.model_output_file, delimiter=args.delimiter, usecols=(args.column,)
    )
    X = np.loadtxt(args.model_input_file, delimiter=args.delimiter, ndmin=2)
    if len(X.shape) == 1:
        X = X.reshape((len(X), 1))

    analyze(
        problem,
        X,
        Y,
        num_resamples=args.resamples,
        print_to_console=True,
        seed=args.seed,
    )


if __name__ == "__main__":
    common_args.run_cli(cli_parse, cli_action)



import numpy as np


def evaluate(values):
    """Linear model (#2) used in Li et al., (2010).

        y = 5x1 + 4x2 + 3x3 + 2x4 + x5

    Parameters
    ----------
    values : np.array


    References
    ----------
    .. [1] Genyuan Li, H. Rabitz, P.E. Yelvington, O.O. Oluwole, F. Bacon,
            C.E. Kolb, and J. Schoendorf, "Global Sensitivity Analysis for
            Systems with Independent and/or Correlated Inputs", Journal of
            Physical Chemistry A, Vol. 114 (19), pp. 6022 - 6032, 2010,
            https://doi.org/10.1021/jp9096919
    """
    Y = np.zeros([values.shape[0]])
    Y = (
        5 * values[:, 0]
        + 4 * values[:, 1]
        + 3 * values[:, 2]
        + 2 * values[:, 3]
        + values[:, 4]
    )

    return Y
